module Bin(Binary(..), putFile, getFile, putBits, getBits) where

import Char
import Control.Exception
import Int(Int32)
import IO
import Time
import Word
import Data.PackedString
--import Zlib

import EIO
import GenUtil(epoch)


binKeyAppend :: Word32 -> Word32 -> Word32
binKeyAppend x y = x + y


putFile :: Binary a => FilePath -> a -> IO ()
putFile  fn x = atomicWrite fn $ \bh -> do
	hSetBuffering bh (BlockBuffering Nothing)
	bh <- makeBitHandle bh
	put bh x
	endBitHandle bh


getFile :: Binary a => FilePath -> IO a
getFile fn = Control.Exception.bracket (openFile fn ReadMode) hClose $ \bh -> do
	hSetBuffering bh (BlockBuffering Nothing)
	bh <- makeBitHandle bh
	get bh 


{-
bufSize :: Int
bufSize = 1024
hGetRawContents :: Handle -> IO [Word8]
hGetRawContents h = do
    a <- newArray_ (0,bufSize)
    getall a where
	getall a = do
	    sz <- hGetArray h a bufSize
	    av <- getElems a
	    if sz == 0 then return [] else do
		r <- getall a
		return (take sz av ++ r)
    
hPutRawContents :: Handle -> [Word8] -> IO ()
hPutRawContents h xs = do
    a <- newArray_ (0,bufSize)
    prc a h xs where
	prc _ _ [] = return ()
	prc a h xs@(_:_) = do
	    let (ys,zs) = splitAt bufSize xs
	    let lys = length ys
	    mapM_ (\(i,e) -> writeArray a i e) (zip [0..lys - 1] ys)
	    hPutArray h a lys
	    prc a h zs
putBitsRawA :: (IORef ([Word8],Int)) -> Int -> Word32 -> IO ()
putBitsRawA _ n _ | n < 0 || n > 32 = error $ "putBits on invalid count: " ++ show n
putBitsRawA _ 0 _ = return ()
putBitsRawA bh nbits val = do
    modifyIORef bh (liftT2 ((([0,1] !! ((fromIntegral val) `mod` 2)):),id))
    putBitsRawA bh (nbits - 1) (val `shiftR` 1)

getBitsRawA :: (IORef ([Word8],Int)) -> Int -> IO Word32
getBitsRawA _ nbits | nbits < 0 || nbits > 32 = error $ "getBits on invalid count: " ++ show nbits
getBitsRawA _ 0 = return 0
getBitsRawA bh nbits = do
    xs <- replicateM nbits (get1 bh)
    return $ foldl (\c x -> (c `shiftL` 1) + x ) 0 (map fromIntegral $ reverse xs) where
	get1 r = do
	    (x:xs,z) <- readIORef r 
	    writeIORef r (xs,z)
	    return x

putBitsRaw :: Handle -> Int -> Word32 -> IO ()
putBitsRaw _ n _ | n < 0 || n > 32 = error $ "putBits on invalid count: " ++ show n
putBitsRaw _ 0 _ = return ()
putBitsRaw bh nbits val = do
    hPutChar bh ("01" !! ((fromIntegral val) `mod` 2))
    putBitsRaw bh (nbits - 1) (val `shiftR` 1)

getBitsRaw :: Handle -> Int -> IO Word32
getBitsRaw _ nbits | nbits < 0 || nbits > 32 = error $ "getBits on invalid count: " ++ show nbits
getBitsRaw _ 0 = return 0
getBitsRaw bh nbits = do
    xs <- replicateM nbits (hGetChar bh)
    let cs = map f $ reverse xs 
    return $ foldl (\c x -> (c `shiftL` 1) + x ) 0 cs where
	f '0' = 0
	f '1' = 1
	f x = error $ "invalid character in savefile: " ++ show x
-}



class Binary a where
    put :: BitHandle -> a -> IO ()
    get :: BitHandle -> IO a
    sizeOf :: a -> Int
    binKey :: a -> Word32
    binKey _ = 0

instance Binary () where
    put  _ () = return ()
    get  _     = return ()
    sizeOf _ = 0
    binKey _ = 1

instance Binary Bool where
    put  bh b = putBits bh 1 (fromIntegral $ fromEnum b)
    get  bh   = getBits bh 1 >>= return . toEnum . fromIntegral
    sizeOf _ = 1
    binKey _ = 2

instance Binary PackedString where
    put bh s = put bh (unpackPS s)
    get bh = get bh >>= return . packString

instance Binary Char where
    put  bh c = put bh (ord c)
    get  bh   = do x <- get bh; return (chr x)
    sizeOf _ = sizeOf (undefined :: Int)
    binKey _ = 3

instance Binary Int where
    put  bh i = putBits bh 32 (fromIntegral i)
    get  bh   = fmap fromIntegral $ getBits bh 32
    sizeOf _ = 32
    binKey _ = 4

instance Binary Word8 where
    put  bh i = putBits bh 8 (fromIntegral i)
    get  bh   = fmap fromIntegral $  getBits bh 8
    sizeOf _ = 8
    binKey _ = 5

instance Binary Int32 where
    put  bh i = putBits bh 32 (fromIntegral i)
    get  bh   = fmap fromIntegral $  getBits bh 32
    sizeOf _ = 32
    binKey _ = 6


instance Binary ClockTime where
    put bh ct = do
	let t = toUTCTime ct
	put bh (ctYear t)
	put bh (fromEnum $ ctMonth t)
	put bh (ctDay t)
	put bh (ctHour t)
	put bh (ctMin t)
	put bh (ctSec t)
    get bh = do
	year <- get bh
	month <- fmap toEnum $ get bh 
	day <- get bh 
	hour <- get bh 
	min <- get bh 
	sec <- get bh 
	return $ toClockTime $ (toUTCTime epoch) {ctYear = year, ctDay = day, ctMonth = month, ctHour = hour, ctMin = min, ctSec = sec}
    sizeOf _ = 32*6
    binKey _ = 100
	

instance (Binary a, Binary b) => Binary (a,b) where
    {-# SPECIALIZE instance (Binary b) => Binary (String, b) #-}
    put bh (a,b) = do put bh a
                      put bh b
    get bh       = do a <- get bh
                      b <- get bh
                      return (a,b)
    sizeOf (a,b) = sizeOf a + sizeOf b
    binKey (a,b) = binKey a `binKeyAppend` binKey b

instance Binary a => Binary [a] where
    {-# SPECIALIZE instance Binary String #-}
    {-# SPECIALIZE instance Binary [Word8] #-}
    put bh []     = putBits bh 1 0
    put bh (x:xs) = do putBits bh 1 1
                       put bh x
                       put bh xs
    get bh        = do h <- getBits bh 1
                       case h of
                          0 -> return []
                          _ -> do x  <- get bh
                                  xs <- get bh
                                  return (x:xs)
    sizeOf xs = sum (map sizeOf xs) + length xs + 1
    binKey (xs::[a]) = binKey (undefined::a) `binKeyAppend` 7

