{-# OPTIONS -#include "Curses_hsc.h" #-}
{-# LINE 1 "Curses.hsc" #-}
module Curses (
{-# LINE 2 "Curses.hsc" #-}
    --------------------------------------------------------------------
    
    Window,      -- data Window deriving Eq
    stdScr,      -- :: Window
    initScr,     -- :: IO Window
    cBreak,      -- :: Bool -> IO ()
    raw,         -- :: Bool -> IO ()
    echo,        -- :: Bool -> IO ()
    nl,          -- :: Bool -> IO ()
    intrFlush,   -- :: Bool -> IO ()
    keypad,      -- :: Window -> Bool -> IO ()
    initCurses,  -- :: IO ()
    endWin,      -- :: IO ()
    resizeTerminal,
    
    --------------------------------------------------------------------
    
    scrSize,          -- :: IO (Int, Int)
    refresh,          -- :: IO ()
    
    --------------------------------------------------------------------
    
    hasColors,      -- :: IO Bool
    startColor,     -- :: IO ()
--    Pair(..),       -- newtype Pair = Pair Int deriving (Eq, Ord, Ix)
--    colorPairs,     -- :: IO Int
    Color(..),      -- newtype Color = Color Int deriving (Eq, Ord, Ix)
    colors,         -- :: IO Int
--    black, red, green, yellow, blue, magenta, cyan, white, -- :: Color
    initPair,       -- :: Pair -> Color -> Color -> IO ()
    pairContent,    -- :: Pair -> IO (Color, Color)
    canChangeColor, -- :: IO Bool
    initColor,      -- :: Color -> (Int, Int, Int) -> IO ()
    colorContent,   -- :: Color -> IO (Int, Int, Int)
    
    --------------------------------------------------------------------
    
    Attr,  -- data Attr deriving Eq
    attr0, -- :: Attr
    
    isAltCharset, isBlink, isBold, isDim, isHorizontal, isInvis,
    isLeft, isLow, isProtect, isReverse, isRight, isStandout, isTop,
    isUnderline, isVertical,
        -- :: Attr -> Bool
    
    setAltCharset, setBlink, setBold, setDim, setHorizontal, setInvis,
    setLeft, setLow, setProtect, setReverse, setRight, setStandout,
    setTop, setUnderline, setVertical,
        -- :: Attr -> Bool -> Attr
    
    attrSet, -- :: Attr -> Pair -> IO ()
    attrOn, attrOff,
    
    --------------------------------------------------------------------

    wAddStr, 
    addLn,         -- :: IO ()
    mvWAddStr,
    wMove,
    
    --------------------------------------------------------------------
    
    bkgrndSet,      -- :: Attr -> Pair -> IO ()
    erase,          -- :: IO ()
    wclear,         -- :: Window -> IO ()
    clrToEol,       -- :: IO ()
    move,           -- :: Int -> Int -> IO ()

    -- Cursor Routines
    CursorVisibility(..), 
    withCursor,

    standout,standend,
    attrDim, attrBold,
    attrDimOn, attrDimOff,
    attrBoldOn, attrBoldOff,
    wAttrOn,
    wAttrOff, 
    touchWin,
    --------------------------------------------------------------------
    -- Mouse Routines
    withMouseEventMask,
    ButtonEvent(..),
    MouseEvent(..),
    
    --------------------------------------------------------------------
    
    Key(..),
    getCh, 
    newPad, pRefresh, delWin, newWin,
    wClrToEol,
    withProgram,

    ulCorner, llCorner, urCorner, lrCorner, rTee, lTee, bTee, tTee, hLine,
    vLine, plus, s1, s9, diamond, ckBoard, degree, plMinus, bullet,
    lArrow, rArrow, dArrow, uArrow, board, lantern, block,
    s3, s7, lEqual, gEqual, pi, nEqual, sterling,

    beep, wAttrSet, wAttrGet,

    cursesSigWinch,
    cursesTest
    ) 
    
    --------------------------------------------------------------------
    where

import Prelude hiding (pi)
import Monad 
import Char           (chr, ord, isPrint, isSpace, toLower)
import Ix             (Ix)

import Data.Bits
import Control.Concurrent
import Foreign
import CForeign

import System.IO.Unsafe
import Control.Exception hiding(block)

import GenUtil(foldl')
import System.Posix.Signals
import CWString
import List 
import Monad
import Maybe



{-# LINE 131 "Curses.hsc" #-}


------------------------------------------------------------------------

fi = fromIntegral

throwIfErr :: Num a => String -> IO a -> IO a
--throwIfErr name act = do
--    res <- act
--    if res == (#const ERR)
--        then ioError (userError ("Curses: "++name++" failed"))
--        else return res
throwIfErr s = throwIf (== (-1)) (\a -> "Curses[" ++ show a ++ "]:"  ++ s)
{-# LINE 144 "Curses.hsc" #-}

throwIfErr_ :: Num a => String -> IO a -> IO ()
throwIfErr_ name act = void $ throwIfErr name act

------------------------------------------------------------------------

data WindowTag 
type Window = Ptr WindowTag

stdScr :: Window
stdScr = unsafePerformIO (peek stdscr)
foreign import ccall "static my_curses.h &stdscr" stdscr :: Ptr Window


initScr :: IO Window
initScr = throwIfNull "initscr" initscr
foreign import ccall unsafe "my_curses.h initscr" initscr :: IO Window

cBreak :: Bool -> IO ()
cBreak True  = throwIfErr_ "cbreak"   cbreak
cBreak False = throwIfErr_ "nocbreak" nocbreak
foreign import ccall unsafe "my_curses.h cbreak" cbreak :: IO CInt
foreign import ccall unsafe "my_curses.h nocbreak" nocbreak :: IO CInt

raw :: Bool -> IO ()
raw False = throwIfErr_ "noraw" noraw
raw True  = throwIfErr_ "raw"   raw_c
foreign import ccall unsafe "my_curses.h noraw" noraw :: IO CInt
foreign import ccall unsafe "my_curses.h raw" raw_c :: IO CInt

echo :: Bool -> IO ()
echo False = throwIfErr_ "noecho" noecho
echo True  = throwIfErr_ "echo"   echo_c
foreign import ccall unsafe "my_curses.h noecho" noecho :: IO CInt
foreign import ccall unsafe "my_curses.h echo" echo_c :: IO CInt

nl :: Bool -> IO ()
nl True  = throwIfErr_ "nl"   nl_c
nl False = throwIfErr_ "nonl" nonl
foreign import ccall unsafe "my_curses.h nl" nl_c :: IO CInt
foreign import ccall unsafe "my_curses.h nonl" nonl :: IO CInt

intrFlush :: Bool -> IO ()
intrFlush bf =
    throwIfErr_ "intrflush" $ intrflush stdScr (if bf then 1 else 0)
foreign import ccall unsafe "my_curses.h intrflush" intrflush :: Window -> (Word8) -> IO CInt
{-# LINE 190 "Curses.hsc" #-}

keypad :: Window -> Bool -> IO ()
keypad win bf =
    throwIfErr_ "keypad" $ keypad_c win (if bf then 1 else 0)
foreign import ccall unsafe "my_curses.h keypad" keypad_c :: Window -> (Word8) -> IO CInt
{-# LINE 195 "Curses.hsc" #-}

foreign import ccall unsafe "my_curses.h leaveok" leaveok_c :: Window -> (Word8) -> IO CInt
{-# LINE 197 "Curses.hsc" #-}

leaveOk True = leaveok_c stdScr 1
leaveOk False = leaveok_c stdScr 0


{-# LINE 202 "Curses.hsc" #-}
foreign import ccall unsafe "my_curses.h use_default_colors" useDefaultColors :: IO ()

defaultBackground = Color (-1)
defaultForeground = Color (-1)


{-# LINE 216 "Curses.hsc" #-}

initCurses :: IO ()
initCurses = do
    initScr
    b <- hasColors
    when b startColor
    --when b useDefaultColors
    cBreak True
    echo False
    nl False
    leaveOk True
    intrFlush False
    keypad stdScr True

endWin :: IO ()
endWin = throwIfErr_ "endwin" endwin
foreign import ccall unsafe "my_curses.h endwin" endwin :: IO CInt

------------------------------------------------------------------------

scrSize :: IO (Int, Int)
scrSize = do
    lines <- peek linesPtr
    cols  <- peek colsPtr
    return (fromIntegral lines, fromIntegral cols)

foreign import ccall "my_curses.h &LINES" linesPtr :: Ptr CInt
foreign import ccall "my_curses.h &COLS" colsPtr :: Ptr CInt


refresh :: IO ()
refresh = throwIfErr_ "refresh" refresh_c
foreign import ccall unsafe "my_curses.h refresh" refresh_c :: IO CInt

------------------------------------------------------------------------

hasColors :: IO Bool
hasColors = liftM (/= 0) has_colors
foreign import ccall unsafe "my_curses.h has_colors" has_colors :: IO (Word8)
{-# LINE 255 "Curses.hsc" #-}

startColor :: IO ()
startColor = throwIfErr_ "start_color" start_color
foreign import ccall unsafe start_color :: IO CInt

newtype Pair = Pair Int deriving (Eq, Ord, Ix)

colorPairs :: IO Int
colorPairs = fmap fromIntegral $ peek colorPairsPtr


foreign import ccall "my_curses.h &COLOR_PAIRS" colorPairsPtr :: Ptr CInt

newtype Color = Color Int deriving (Eq, Ord, Ix)

colors :: IO Int
colors = liftM fromIntegral $ peek colorsPtr

foreign import ccall "my_curses.h &COLORS" colorsPtr :: Ptr CInt

--black, red, green, yellow, blue, magenta, cyan, white :: Color

color :: String -> Maybe Color
color "default"  = Just $ Color (-1)
color "black"    = Just $ Color (0)
{-# LINE 280 "Curses.hsc" #-}
color "red"      = Just $ Color (1)
{-# LINE 281 "Curses.hsc" #-}
color "green"    = Just $ Color (2)
{-# LINE 282 "Curses.hsc" #-}
color "yellow"   = Just $ Color (3)
{-# LINE 283 "Curses.hsc" #-}
color "blue"     = Just $ Color (4)
{-# LINE 284 "Curses.hsc" #-}
color "magenta"  = Just $ Color (5)
{-# LINE 285 "Curses.hsc" #-}
color "cyan"     = Just $ Color (6)
{-# LINE 286 "Curses.hsc" #-}
color "white"    = Just $ Color (7)
{-# LINE 287 "Curses.hsc" #-}
color _ =  Nothing

data Attribute = Attribute [String] String String
parseAttr :: String -> Attribute 
parseAttr s = Attribute as fg bg where
    rs = filter (not . f . head) $ groupBy (\x y -> f x && f y) (map toLower s) 
    as = filter (`elem` attributes) rs
    col x = if isJust (color x) then return x else Nothing
    fg = fromJust $ msum (map (cGet "fg") rs)  `mplus` msum (map col rs) `mplus` return "default"
    bg = fromJust $ msum (map (cGet "bg") rs) `mplus` return "default"
    f ',' = True
    f c | isSpace c = True
    f _ = False
    cGet p r | (p ++ ":") `isPrefixOf` r = col (drop (length p + 1) r) 
    cGet _ _ = Nothing
    attributes = ["normal", "bold", "blink", "dim", "reverse", "underline" ]


initPair :: Pair -> Color -> Color -> IO ()
initPair (Pair p) (Color f) (Color b) =
    throwIfErr_ "init_pair" $
        init_pair (fromIntegral p) (fromIntegral f) (fromIntegral b)
foreign import ccall unsafe init_pair :: CShort -> CShort -> CShort -> IO CInt

pairContent :: Pair -> IO (Color, Color)
pairContent (Pair p) =
    alloca $ \fPtr ->
    alloca $ \bPtr -> do
        throwIfErr "pair_content" $ pair_content (fromIntegral p) fPtr bPtr
        f <- peek fPtr
        b <- peek bPtr
        return (Color (fromIntegral f), Color (fromIntegral b))
foreign import ccall unsafe pair_content :: CShort -> Ptr CShort -> Ptr CShort -> IO CInt

canChangeColor :: IO Bool
canChangeColor = liftM (/= 0) can_change_color
foreign import ccall unsafe can_change_color :: IO (Word8)
{-# LINE 324 "Curses.hsc" #-}

initColor :: Color -> (Int, Int, Int) -> IO ()
initColor (Color c) (r, g, b) =
    throwIfErr_ "init_color" $
        init_color (fromIntegral c) (fromIntegral r) (fromIntegral g) (fromIntegral b)
foreign import ccall unsafe init_color :: CShort -> CShort -> CShort -> CShort -> IO CInt

colorContent :: Color -> IO (Int, Int, Int)
colorContent (Color c) =
    alloca $ \rPtr ->
    alloca $ \gPtr ->
    alloca $ \bPtr -> do
        throwIfErr "color_content" $ color_content (fromIntegral c) rPtr gPtr bPtr
        r <- peek rPtr
        g <- peek gPtr
        b <- peek bPtr
        return (fromIntegral r, fromIntegral g, fromIntegral b)
foreign import ccall unsafe color_content :: CShort -> Ptr CShort -> Ptr CShort -> Ptr CShort -> IO CInt

foreign import ccall unsafe "my_curses.h hs_curses_color_pair" colorPair :: Pair -> (Word32)
{-# LINE 344 "Curses.hsc" #-}

{-# LINE 345 "Curses.hsc" #-}

-------------
-- Attributes 
-------------

foreign import ccall unsafe "my_curses.h attr_set" attr_set :: Attr -> CShort -> Ptr a -> IO Int
-- foreign import ccall unsafe "my_curses.h attr_get" :: Attr -> CShort -> Ptr a -> IO Int

foreign import ccall unsafe "my_curses.h wattr_set" wattr_set :: Window -> Attr -> CInt -> Ptr a -> IO CInt
foreign import ccall unsafe "my_curses.h wattr_get" wattr_get :: Window -> Ptr Attr -> Ptr CShort -> Ptr a -> IO CInt

foreign import ccall "my_curses.h attr_on" attr_on :: (Word32) -> Ptr a -> IO Int
{-# LINE 357 "Curses.hsc" #-}
foreign import ccall "my_curses.h attr_off" attr_off :: (Word32) -> Ptr a -> IO Int
{-# LINE 358 "Curses.hsc" #-}
foreign import ccall "my_curses.h attron" attron :: Int -> IO Int
foreign import ccall "my_curses.h attroff" attroff :: Int -> IO Int
foreign import ccall unsafe "my_curses.h wattron" wattron :: Window -> CInt -> IO CInt
foreign import ccall unsafe "my_curses.h wattroff" wattroff :: Window -> CInt -> IO CInt
foreign import ccall standout :: IO Int
foreign import ccall standend :: IO Int

wAttrSet :: Window -> (Attr,Pair) -> IO ()
wAttrSet w (a,(Pair p)) = throwIfErr_ "wattr_set" $ wattr_set w a (fromIntegral p) nullPtr
wAttrGet :: Window -> IO (Attr,Pair)
wAttrGet w =  alloca $ \pa -> alloca $ \pp -> (throwIfErr_ "wattr_get" $ wattr_get w pa pp nullPtr) >> (peek pa >>=  \a -> peek pp >>=  \p -> return (a,Pair (fromIntegral p)))


newtype Attr = Attr (Word32) deriving (Eq,Storable,Bits, Num, Show)
{-# LINE 372 "Curses.hsc" #-}

attr0 :: Attr
attr0 = Attr (0)
{-# LINE 375 "Curses.hsc" #-}

isAltCharset, isBlink, isBold, isDim, isHorizontal, isInvis, isLeft,
    isLow, isProtect, isReverse, isRight, isStandout, isTop,
    isUnderline, isVertical
    :: Attr -> Bool
isAltCharset = isAttr (4194304)
{-# LINE 381 "Curses.hsc" #-}
isBlink      = isAttr (524288)
{-# LINE 382 "Curses.hsc" #-}
isBold       = isAttr (2097152)
{-# LINE 383 "Curses.hsc" #-}
isDim        = isAttr (1048576)
{-# LINE 384 "Curses.hsc" #-}
isHorizontal = isAttr (33554432)
{-# LINE 385 "Curses.hsc" #-}
isInvis      = isAttr (8388608)
{-# LINE 386 "Curses.hsc" #-}
isLeft       = isAttr (67108864)
{-# LINE 387 "Curses.hsc" #-}
isLow        = isAttr (134217728)
{-# LINE 388 "Curses.hsc" #-}
isProtect    = isAttr (16777216)
{-# LINE 389 "Curses.hsc" #-}
isReverse    = isAttr (262144)
{-# LINE 390 "Curses.hsc" #-}
isRight      = isAttr (268435456)
{-# LINE 391 "Curses.hsc" #-}
isStandout   = isAttr (65536)
{-# LINE 392 "Curses.hsc" #-}
isTop        = isAttr (536870912)
{-# LINE 393 "Curses.hsc" #-}
isUnderline  = isAttr (131072)
{-# LINE 394 "Curses.hsc" #-}
isVertical   = isAttr (1073741824)
{-# LINE 395 "Curses.hsc" #-}

isAttr :: (Word32) -> Attr -> Bool
{-# LINE 397 "Curses.hsc" #-}
isAttr bit (Attr a) = a .&. bit /= 0

setAltCharset, setBlink, setBold, setDim, setHorizontal, setInvis,
    setLeft, setLow, setProtect, setReverse, setRight, setStandout,
    setTop, setUnderline, setVertical
    :: Attr -> Bool -> Attr
setAltCharset = setAttr (4194304)
{-# LINE 404 "Curses.hsc" #-}
setBlink      = setAttr (524288)
{-# LINE 405 "Curses.hsc" #-}
setBold       = setAttr (2097152)
{-# LINE 406 "Curses.hsc" #-}
setDim        = setAttr (1048576)
{-# LINE 407 "Curses.hsc" #-}
setHorizontal = setAttr (33554432)
{-# LINE 408 "Curses.hsc" #-}
setInvis      = setAttr (8388608)
{-# LINE 409 "Curses.hsc" #-}
setLeft       = setAttr (67108864)
{-# LINE 410 "Curses.hsc" #-}
setLow        = setAttr (134217728)
{-# LINE 411 "Curses.hsc" #-}
setProtect    = setAttr (16777216)
{-# LINE 412 "Curses.hsc" #-}
setReverse    = setAttr (262144)
{-# LINE 413 "Curses.hsc" #-}
setRight      = setAttr (268435456)
{-# LINE 414 "Curses.hsc" #-}
setStandout   = setAttr (65536)
{-# LINE 415 "Curses.hsc" #-}
setTop        = setAttr (536870912)
{-# LINE 416 "Curses.hsc" #-}
setUnderline  = setAttr (131072)
{-# LINE 417 "Curses.hsc" #-}
setVertical   = setAttr (1073741824)
{-# LINE 418 "Curses.hsc" #-}

setAttr :: (Word32) -> Attr -> Bool -> Attr
{-# LINE 420 "Curses.hsc" #-}
setAttr bit (Attr a) False = Attr (a .&. complement bit)
setAttr bit (Attr a) True  = Attr (a .|.            bit)

attrSet :: Attr -> Pair -> IO ()
attrSet attr (Pair p) = throwIfErr_ "attrset" $
    attr_set attr (fromIntegral p) nullPtr

attrOn :: Attr -> IO ()
attrOn (Attr attr) = throwIfErr_ "attr_on" $
    attr_on attr nullPtr


attrOff :: Attr -> IO ()
attrOff (Attr attr) = throwIfErr_ "attr_off" $
    attr_off attr nullPtr



wAttrOn :: Window -> Int -> IO ()
wAttrOn w x = throwIfErr_ "wattron" $ wattron w (fi x)

wAttrOff :: Window -> Int -> IO ()
wAttrOff w x = throwIfErr_ "wattroff" $ wattroff w (fi x)

attrDimOn :: IO ()
attrDimOn  = throwIfErr_ "attron A_DIM" $
    attron (1048576) 
{-# LINE 447 "Curses.hsc" #-}

attrDimOff :: IO ()
attrDimOff = throwIfErr_ "attroff A_DIM" $
    attroff (1048576) 
{-# LINE 451 "Curses.hsc" #-}

attrBoldOn :: IO ()
attrBoldOn  = throwIfErr_ "attron A_BOLD" $
    attron (2097152) 
{-# LINE 455 "Curses.hsc" #-}

attrBoldOff :: IO ()
attrBoldOff = throwIfErr_ "attroff A_BOLD" $
    attroff (2097152) 
{-# LINE 459 "Curses.hsc" #-}


attrDim :: Int
attrDim = (1048576)
{-# LINE 463 "Curses.hsc" #-}
attrBold :: Int
attrBold = (2097152)
{-# LINE 465 "Curses.hsc" #-}

------------------------------------------------------------------------


mvWAddStr :: Window -> Int -> Int -> String -> IO ()
mvWAddStr w y x str = wMove w y x >> wAddStr w str 

addLn :: IO ()
addLn = wAddStr stdScr "\n" 

sanifyOutput :: String -> String
sanifyOutput = map f . filter (/= '\r') where
    f c | isPrint c  = c
    f c = '~'


{-# LINE 488 "Curses.hsc" #-}

wAddStr :: Window -> String -> IO ()
wAddStr w str =  withLCStringLen (sanifyOutput str) (\(ws,len) -> throwIfErr_ ("waddnstr: " ++ show len ++ " " ++ show str) $ waddnstr w ws (fi len))
    
foreign import ccall unsafe waddnstr :: Window -> CString -> CInt -> IO CInt


{-# LINE 495 "Curses.hsc" #-}

------------------------------------------------------------------------


{-# LINE 501 "Curses.hsc" #-}

bkgrndSet :: Attr -> Pair -> IO ()
bkgrndSet (Attr a) p = bkgdset $
    fromIntegral (ord ' ') .|.
    (if a .&. 4194304 /= 0 then 4194304 else 0) .|.
{-# LINE 506 "Curses.hsc" #-}
    (if a .&. 524288 /= 0 then 524288 else 0) .|.
{-# LINE 507 "Curses.hsc" #-}
    (if a .&. 2097152 /= 0 then 2097152 else 0) .|.
{-# LINE 508 "Curses.hsc" #-}
    (if a .&. 1048576 /= 0 then 1048576 else 0) .|.
{-# LINE 509 "Curses.hsc" #-}
    (if a .&. 8388608 /= 0 then 8388608 else 0) .|.
{-# LINE 510 "Curses.hsc" #-}
    (if a .&. 16777216 /= 0 then 16777216 else 0) .|.
{-# LINE 511 "Curses.hsc" #-}
    (if a .&. 262144 /= 0 then 262144 else 0) .|.
{-# LINE 512 "Curses.hsc" #-}
    (if a .&. 65536 /= 0 then 65536 else 0) .|.
{-# LINE 513 "Curses.hsc" #-}
    (if a .&. 131072 /= 0 then 131072 else 0) .|.
{-# LINE 514 "Curses.hsc" #-}
    colorPair p
foreign import ccall unsafe bkgdset :: (Word32) -> IO ()
{-# LINE 516 "Curses.hsc" #-}

erase :: IO ()
erase = throwIfErr_ "erase" $ werase_c  stdScr
foreign import ccall unsafe "werase" werase_c :: Window -> IO CInt

wclear :: Window -> IO ()
wclear w = throwIfErr_ "wclear" $ wclear_c  w
foreign import ccall unsafe "wclear" wclear_c :: Window -> IO CInt



clrToEol :: IO ()
clrToEol = throwIfErr_ "clrtoeol" clrtoeol
foreign import ccall unsafe clrtoeol :: IO CInt

move :: Int -> Int -> IO ()
move y x =
    throwIfErr_ "move" $ move_c (fromIntegral y) (fromIntegral x)
foreign import ccall unsafe "move" move_c :: CInt -> CInt -> IO CInt

wMove :: Window -> Int -> Int -> IO ()
wMove w y x = throwIfErr_ "wmove" $ wmove w (fi y) (fi x)
foreign import ccall unsafe  wmove :: Window -> CInt -> CInt -> IO CInt

------------------
-- Cursor routines
------------------

data CursorVisibility = CursorInvisible | CursorVisible | CursorVeryVisible

vis_c :: CursorVisibility -> CInt
vis_c vis = case vis of
    CursorInvisible   -> 0
    CursorVisible     -> 1
    CursorVeryVisible -> 2
    
foreign import ccall unsafe "my_curses.h curs_set" curs_set :: CInt -> IO CInt

cursSet 0 = leaveOk True >> curs_set 0
cursSet n = leaveOk False >> curs_set n 

withCursor :: CursorVisibility -> IO a -> IO a
withCursor nv action = Control.Exception.bracket (cursSet (vis_c nv)) (\v -> case v of 
		(-1) -> return 0
{-# LINE 560 "Curses.hsc" #-}
		x -> cursSet x) (\_ -> action)



------------------------------------------------------------------------


touchWin :: Window -> IO ()
touchWin w = throwIfErr_ "touchwin" $ touchwin w
foreign import ccall touchwin :: Window -> IO CInt

newPad :: Int -> Int -> IO Window
newPad nlines ncols = throwIfNull "newpad" $ newpad (fromIntegral nlines) (fromIntegral ncols)

pRefresh :: Window -> Int -> Int -> Int -> Int -> Int -> Int -> IO ()
pRefresh pad pminrow pmincol sminrow smincol smaxrow smaxcol = throwIfErr_ "prefresh" $
    prefresh pad (fromIntegral pminrow) (fromIntegral pmincol) (fromIntegral sminrow) (fromIntegral smincol) (fromIntegral smaxrow) (fromIntegral smaxcol)

delWin :: Window -> IO ()
delWin w = throwIfErr_ "delwin" $ delwin w
    
foreign import ccall unsafe prefresh :: Window -> CInt -> CInt -> CInt -> CInt -> CInt -> CInt -> IO CInt
foreign import ccall unsafe newpad :: CInt -> CInt -> IO Window
foreign import ccall unsafe delwin :: Window -> IO CInt


newWin :: Int -> Int -> Int -> Int -> IO Window
newWin nlines ncolumn begin_y begin_x = throwIfNull "newwin" $ newwin (fi nlines) (fi ncolumn) (fi begin_y) (fi begin_x)

foreign import ccall unsafe newwin :: CInt -> CInt -> CInt -> CInt -> IO Window


wClrToEol :: Window -> IO ()
wClrToEol w = throwIfErr_ "wclrtoeol" $ wclrtoeol w
foreign import ccall unsafe wclrtoeol :: Window -> IO CInt




foreign import ccall threadsafe getch :: IO CInt


--foreign import ccall unsafe def_prog_mode :: IO CInt
--foreign import ccall unsafe reset_prog_mode :: IO CInt
foreign import ccall unsafe flushinp :: IO CInt


withProgram :: IO a -> IO a
withProgram action = withCursor CursorVisible $ Control.Exception.bracket_ (endWin) (flushinp) action
--withProgram action = withCursor CursorVisible $ Control.Exception.bracket_ ({-def_prog_mode >> -}endWin) (return ()){-reset_prog_mode-} action


foreign import ccall unsafe "my_curses.h beep" c_beep :: IO CInt
foreign import ccall unsafe "my_curses.h flash" c_flash :: IO CInt

beep :: IO ()
beep = do
    br <- c_beep
    when (br /= (0)) (c_flash >> return ()) 
{-# LINE 619 "Curses.hsc" #-}


---------------
-- Key Routines
---------------

data Key
    = KeyChar Char | KeyBreak | KeyDown | KeyUp | KeyLeft | KeyRight
    | KeyHome | KeyBackspace | KeyF Int | KeyDL | KeyIL | KeyDC
    | KeyIC | KeyEIC | KeyClear | KeyEOS | KeyEOL | KeySF | KeySR
    | KeyNPage | KeyPPage | KeySTab | KeyCTab | KeyCATab | KeyEnter
    | KeySReset | KeyReset | KeyPrint | KeyLL | KeyA1 | KeyA3
    | KeyB2 | KeyC1 | KeyC3 | KeyBTab | KeyBeg | KeyCancel | KeyClose
    | KeyCommand | KeyCopy | KeyCreate | KeyEnd | KeyExit | KeyFind
    | KeyHelp | KeyMark | KeyMessage | KeyMove | KeyNext | KeyOpen
    | KeyOptions | KeyPrevious | KeyRedo | KeyReference | KeyRefresh
    | KeyReplace | KeyRestart | KeyResume | KeySave | KeySBeg
    | KeySCancel | KeySCommand | KeySCopy | KeySCreate | KeySDC
    | KeySDL | KeySelect | KeySEnd | KeySEOL | KeySExit | KeySFind
    | KeySHelp | KeySHome | KeySIC | KeySLeft | KeySMessage | KeySMove
    | KeySNext | KeySOptions | KeySPrevious | KeySPrint | KeySRedo
    | KeySReplace | KeySRight | KeySRsume | KeySSave | KeySSuspend
    | KeySUndo | KeySuspend | KeyUndo | KeyResize | KeyMouse | KeyUnknown Int
    deriving (Eq,Show,Ord)

decodeKey :: CInt -> Key
decodeKey key = case key of
    _ | key >= 0 && key <= 255 -> KeyChar (chr (fromIntegral key))
    (257)         -> KeyBreak
{-# LINE 648 "Curses.hsc" #-}
    (258)          -> KeyDown
{-# LINE 649 "Curses.hsc" #-}
    (259)            -> KeyUp
{-# LINE 650 "Curses.hsc" #-}
    (260)          -> KeyLeft
{-# LINE 651 "Curses.hsc" #-}
    (261)         -> KeyRight
{-# LINE 652 "Curses.hsc" #-}
    (262)          -> KeyHome
{-# LINE 653 "Curses.hsc" #-}
    (263)     -> KeyBackspace
{-# LINE 654 "Curses.hsc" #-}
    _ | key >= (264) && key <= (327)
{-# LINE 655 "Curses.hsc" #-}
                               -> KeyF (fromIntegral (key - 264))
{-# LINE 656 "Curses.hsc" #-}
    (328)            -> KeyDL
{-# LINE 657 "Curses.hsc" #-}
    (329)            -> KeyIL
{-# LINE 658 "Curses.hsc" #-}
    (330)            -> KeyDC
{-# LINE 659 "Curses.hsc" #-}
    (331)            -> KeyIC
{-# LINE 660 "Curses.hsc" #-}
    (332)           -> KeyEIC
{-# LINE 661 "Curses.hsc" #-}
    (333)         -> KeyClear
{-# LINE 662 "Curses.hsc" #-}
    (334)           -> KeyEOS
{-# LINE 663 "Curses.hsc" #-}
    (335)           -> KeyEOL
{-# LINE 664 "Curses.hsc" #-}
    (336)            -> KeySF
{-# LINE 665 "Curses.hsc" #-}
    (337)            -> KeySR
{-# LINE 666 "Curses.hsc" #-}
    (338)         -> KeyNPage
{-# LINE 667 "Curses.hsc" #-}
    (339)         -> KeyPPage
{-# LINE 668 "Curses.hsc" #-}
    (340)          -> KeySTab
{-# LINE 669 "Curses.hsc" #-}
    (341)          -> KeyCTab
{-# LINE 670 "Curses.hsc" #-}
    (342)         -> KeyCATab
{-# LINE 671 "Curses.hsc" #-}
    (343)         -> KeyEnter
{-# LINE 672 "Curses.hsc" #-}
    (344)        -> KeySReset
{-# LINE 673 "Curses.hsc" #-}
    (345)         -> KeyReset
{-# LINE 674 "Curses.hsc" #-}
    (346)         -> KeyPrint
{-# LINE 675 "Curses.hsc" #-}
    (347)            -> KeyLL
{-# LINE 676 "Curses.hsc" #-}
    (348)            -> KeyA1
{-# LINE 677 "Curses.hsc" #-}
    (349)            -> KeyA3
{-# LINE 678 "Curses.hsc" #-}
    (350)            -> KeyB2
{-# LINE 679 "Curses.hsc" #-}
    (351)            -> KeyC1
{-# LINE 680 "Curses.hsc" #-}
    (352)            -> KeyC3
{-# LINE 681 "Curses.hsc" #-}
    (353)          -> KeyBTab
{-# LINE 682 "Curses.hsc" #-}
    (354)           -> KeyBeg
{-# LINE 683 "Curses.hsc" #-}
    (355)        -> KeyCancel
{-# LINE 684 "Curses.hsc" #-}
    (356)         -> KeyClose
{-# LINE 685 "Curses.hsc" #-}
    (357)       -> KeyCommand
{-# LINE 686 "Curses.hsc" #-}
    (358)          -> KeyCopy
{-# LINE 687 "Curses.hsc" #-}
    (359)        -> KeyCreate
{-# LINE 688 "Curses.hsc" #-}
    (360)           -> KeyEnd
{-# LINE 689 "Curses.hsc" #-}
    (361)          -> KeyExit
{-# LINE 690 "Curses.hsc" #-}
    (362)          -> KeyFind
{-# LINE 691 "Curses.hsc" #-}
    (363)          -> KeyHelp
{-# LINE 692 "Curses.hsc" #-}
    (364)          -> KeyMark
{-# LINE 693 "Curses.hsc" #-}
    (365)       -> KeyMessage
{-# LINE 694 "Curses.hsc" #-}
    (366)          -> KeyMove
{-# LINE 695 "Curses.hsc" #-}
    (367)          -> KeyNext
{-# LINE 696 "Curses.hsc" #-}
    (368)          -> KeyOpen
{-# LINE 697 "Curses.hsc" #-}
    (369)       -> KeyOptions
{-# LINE 698 "Curses.hsc" #-}
    (370)      -> KeyPrevious
{-# LINE 699 "Curses.hsc" #-}
    (371)          -> KeyRedo
{-# LINE 700 "Curses.hsc" #-}
    (372)     -> KeyReference
{-# LINE 701 "Curses.hsc" #-}
    (373)       -> KeyRefresh
{-# LINE 702 "Curses.hsc" #-}
    (374)       -> KeyReplace
{-# LINE 703 "Curses.hsc" #-}
    (375)       -> KeyRestart
{-# LINE 704 "Curses.hsc" #-}
    (376)        -> KeyResume
{-# LINE 705 "Curses.hsc" #-}
    (377)          -> KeySave
{-# LINE 706 "Curses.hsc" #-}
    (378)          -> KeySBeg
{-# LINE 707 "Curses.hsc" #-}
    (379)       -> KeySCancel
{-# LINE 708 "Curses.hsc" #-}
    (380)      -> KeySCommand
{-# LINE 709 "Curses.hsc" #-}
    (381)         -> KeySCopy
{-# LINE 710 "Curses.hsc" #-}
    (382)       -> KeySCreate
{-# LINE 711 "Curses.hsc" #-}
    (383)           -> KeySDC
{-# LINE 712 "Curses.hsc" #-}
    (384)           -> KeySDL
{-# LINE 713 "Curses.hsc" #-}
    (385)        -> KeySelect
{-# LINE 714 "Curses.hsc" #-}
    (386)          -> KeySEnd
{-# LINE 715 "Curses.hsc" #-}
    (387)          -> KeySEOL
{-# LINE 716 "Curses.hsc" #-}
    (388)         -> KeySExit
{-# LINE 717 "Curses.hsc" #-}
    (389)         -> KeySFind
{-# LINE 718 "Curses.hsc" #-}
    (390)         -> KeySHelp
{-# LINE 719 "Curses.hsc" #-}
    (391)         -> KeySHome
{-# LINE 720 "Curses.hsc" #-}
    (392)           -> KeySIC
{-# LINE 721 "Curses.hsc" #-}
    (393)         -> KeySLeft
{-# LINE 722 "Curses.hsc" #-}
    (394)      -> KeySMessage
{-# LINE 723 "Curses.hsc" #-}
    (395)         -> KeySMove
{-# LINE 724 "Curses.hsc" #-}
    (396)         -> KeySNext
{-# LINE 725 "Curses.hsc" #-}
    (397)      -> KeySOptions
{-# LINE 726 "Curses.hsc" #-}
    (398)     -> KeySPrevious
{-# LINE 727 "Curses.hsc" #-}
    (399)        -> KeySPrint
{-# LINE 728 "Curses.hsc" #-}
    (400)         -> KeySRedo
{-# LINE 729 "Curses.hsc" #-}
    (401)      -> KeySReplace
{-# LINE 730 "Curses.hsc" #-}
    (402)        -> KeySRight
{-# LINE 731 "Curses.hsc" #-}
    (403)        -> KeySRsume
{-# LINE 732 "Curses.hsc" #-}
    (404)         -> KeySSave
{-# LINE 733 "Curses.hsc" #-}
    (405)      -> KeySSuspend
{-# LINE 734 "Curses.hsc" #-}
    (406)         -> KeySUndo
{-# LINE 735 "Curses.hsc" #-}
    (407)       -> KeySuspend
{-# LINE 736 "Curses.hsc" #-}
    (408)          -> KeyUndo
{-# LINE 737 "Curses.hsc" #-}

{-# LINE 738 "Curses.hsc" #-}
    (410)        -> KeyResize
{-# LINE 739 "Curses.hsc" #-}

{-# LINE 740 "Curses.hsc" #-}

{-# LINE 741 "Curses.hsc" #-}
    (409)        -> KeyMouse
{-# LINE 742 "Curses.hsc" #-}

{-# LINE 743 "Curses.hsc" #-}
    _                          -> KeyUnknown (fromIntegral key)





getCh :: IO Key
getCh = threadWaitRead 0 >> (liftM decodeKey $ throwIfErr "getch" getch)

--getCh :: IO Key
--getCh = liftM decodeKey $ throwIfErr "getch" getch

-- getCh :: IO Key
-- getCh = do
--     nodelay stdScr 1
--     --halfdelay 1
--     v <- getch
--     case v of
-- 	(#const ERR) -> yield >> getCh 
-- 	x -> return $ decodeKey x


resizeTerminal :: Int -> Int -> IO ()


{-# LINE 768 "Curses.hsc" #-}

resizeTerminal a b = throwIfErr_ "resizeterm"  $ resizeterm (fi a) (fi b)

foreign import ccall unsafe "my_curses.h resizeterm" resizeterm :: CInt -> CInt -> IO CInt


{-# LINE 778 "Curses.hsc" #-}


cursesSigWinch :: Maybe Signal


{-# LINE 783 "Curses.hsc" #-}

cursesSigWinch = Just (28)
{-# LINE 785 "Curses.hsc" #-}


{-# LINE 791 "Curses.hsc" #-}



------------
-- Test case
------------

cursesTest :: IO ()
cursesTest = do
    initScr
    hc <- hasColors 
    when hc startColor
    ccc <- canChangeColor
    (ys,xs) <- scrSize
    cp <- colorPairs
    cs <- colors
    endWin
    putStrLn $ "ScreenSize: " ++ show (xs,ys) 
    putStrLn $ "hasColors: " ++ show hc
    putStrLn $ "canChangeColor: " ++ show ccc
    putStrLn $ "colorPairs: " ++ show cp
    putStrLn $ "colors: " ++ show cs

    


-----------------
-- Mouse Routines
-----------------

data MouseEvent = MouseEvent {
    mouseEventId :: Int, 
    mouseEventX :: Int, 
    mouseEventY :: Int, 
    mouseEventZ :: Int, 
    mouseEventButton :: [ButtonEvent]
   } deriving(Show)

data ButtonEvent = ButtonPressed Int | ButtonReleased Int | ButtonClicked Int | 
    ButtonDoubleClicked Int | ButtonTripleClicked Int | ButtonShift | ButtonControl | ButtonAlt 
	deriving(Eq,Show)

withMouseEventMask :: [ButtonEvent] -> IO a -> IO a


{-# LINE 836 "Curses.hsc" #-}

foreign import ccall unsafe "my_curses.h mousemask" mousemask :: (Word32) -> Ptr (Word32) -> IO (Word32)
{-# LINE 838 "Curses.hsc" #-}

withMouseEventMask bes action = do
    ov <- alloca (\a ->  mousemask (besToMouseMask bes) a >> peek a) 
    r <- action 
    mousemask ov nullPtr 
    return r

besToMouseMask :: [ButtonEvent] -> (Word32)
{-# LINE 846 "Curses.hsc" #-}
besToMouseMask bes = foldl' (.|.) 0 (map cb bes) where
    cb (ButtonPressed 1) = (2)
{-# LINE 848 "Curses.hsc" #-}
    cb (ButtonPressed 2) = (128)
{-# LINE 849 "Curses.hsc" #-}
    cb (ButtonPressed 3) = (8192)
{-# LINE 850 "Curses.hsc" #-}
    cb (ButtonPressed 4) = (524288)
{-# LINE 851 "Curses.hsc" #-}
    cb (ButtonReleased 1) = (1)
{-# LINE 852 "Curses.hsc" #-}
    cb (ButtonReleased 2) = (64)
{-# LINE 853 "Curses.hsc" #-}
    cb (ButtonReleased 3) = (4096)
{-# LINE 854 "Curses.hsc" #-}
    cb (ButtonReleased 4) = (262144)
{-# LINE 855 "Curses.hsc" #-}
    cb (ButtonClicked 1) = (4)
{-# LINE 856 "Curses.hsc" #-}
    cb (ButtonClicked 2) = (256)
{-# LINE 857 "Curses.hsc" #-}
    cb (ButtonClicked 3) = (16384)
{-# LINE 858 "Curses.hsc" #-}
    cb (ButtonClicked 4) = (1048576)
{-# LINE 859 "Curses.hsc" #-}
    cb ButtonShift = (33554432)
{-# LINE 860 "Curses.hsc" #-}
    cb ButtonAlt = (67108864)
{-# LINE 861 "Curses.hsc" #-}
    cb ButtonControl = (16777216)
{-# LINE 862 "Curses.hsc" #-}
    cb _ = 0



{-# LINE 869 "Curses.hsc" #-}


-------------------------
-- code graveyard
-------------------------








ulCorner, llCorner, urCorner, lrCorner, rTee, lTee, bTee, tTee, hLine,
    vLine, plus, s1, s9, diamond, ckBoard, degree, plMinus, bullet,
    lArrow, rArrow, dArrow, uArrow, board, lantern, block,
    s3, s7, lEqual, gEqual, pi, nEqual, sterling
    :: Char

ulCorner = chr 0x250C
llCorner = chr 0x2514
urCorner = chr 0x2510
lrCorner = chr 0x2518
rTee     = chr 0x2524
lTee     = chr 0x251C
bTee     = chr 0x2534
tTee     = chr 0x252C
hLine    = chr 0x2500
vLine    = chr 0x2502
plus     = chr 0x253C
s1       = chr 0x23BA -- was: 0xF800
s9       = chr 0x23BD -- was: 0xF804
diamond  = chr 0x25C6
ckBoard  = chr 0x2592
degree   = chr 0x00B0
plMinus  = chr 0x00B1
bullet   = chr 0x00B7
lArrow   = chr 0x2190
rArrow   = chr 0x2192
dArrow   = chr 0x2193
uArrow   = chr 0x2191
board    = chr 0x2591
lantern  = chr 0x256C
block    = chr 0x2588
s3       = chr 0x23BB -- was: 0xF801
s7       = chr 0x23BC -- was: 0xF803
lEqual   = chr 0x2264
gEqual   = chr 0x2265
pi       = chr 0x03C0
nEqual   = chr 0x2260
sterling = chr 0x00A3


{-# LINE 1072 "Curses.hsc" #-}
